from __future__ import annotations

import re
import typing as t

import pydantic

from taktile_auth.enums import Wildcard

PARTIAL_RE = re.compile(r"(^[^\*]*$)|(^[^\*]*\*$)")


def is_full_wildcard(v: str) -> str:
    assert v == "*" or "*" not in v
    return v


def is_partial_wildcard(v: str) -> str:
    assert re.fullmatch(PARTIAL_RE, v)
    return v


def is_fully_specified(v: str) -> str:
    assert "*" not in v
    return v


WILDCARD_CHECK = {
    Wildcard.allowed: is_full_wildcard,
    Wildcard.partial: is_partial_wildcard,
    Wildcard.not_allowed: is_fully_specified,
}


class Resource(pydantic.BaseModel):
    def __contains__(self, o: Resource) -> bool:
        """Checks if the queried resource 'o' is either equal
        or contained inside of 'self'.
        For each arg of the resource, checks if it is either a
        match with query's corresponding arg (including wildcards).
        """

        if self.schema()["title"] != o.schema()["title"]:
            return False

        for arg in self.dict().keys():
            allowed = getattr(self, arg)
            queried = getattr(o, arg)
            if not re.fullmatch(allowed.replace("*", ".*"), queried):
                return False
        return True

    def __hash__(self) -> int:
        return hash((type(self).__name__,) + tuple(self.__dict__.values()))


class ResourceDefinition(pydantic.BaseModel):
    resource_name: str
    args: t.Dict[str, Wildcard]

    def get_resource(self) -> t.Type[Resource]:
        fields: t.Any = {
            field_name: (str, ...) for field_name in self.args.keys()
        }
        validators = {
            f"{field_name}_validator": (
                pydantic.validator(field_name, allow_reuse=True)(
                    WILDCARD_CHECK[check]
                )
            )
            for field_name, check in self.args.items()
        }
        return t.cast(
            t.Type[Resource],
            pydantic.create_model(
                self.resource_name,
                **fields,
                __validators__=validators,
                __base__=Resource,
            ),
        )



import json
import typing as t

import jwt
import requests
from jwt.algorithms import RSAAlgorithm
from pydantic import ValidationError

from taktile_auth.exceptions import InvalidAuthException
from taktile_auth.schemas.session import SessionState
from taktile_auth.schemas.token import TaktileIdToken
from taktile_auth.settings import settings

ALGORITHM = "RS256"


def _get_auth_server_url(env: str) -> str:
    if env == "local":
        return "http://taktile-api.local.taktile.com:8000"
    if env == "prod":
        return "https://taktile-api.taktile.com"
    return f"https://taktile-api.{env}.taktile.com"


class AuthClient:
    def __init__(
        self,
        url: str = _get_auth_server_url(settings.ENV),
    ) -> None:
        self.public_key_url = f"{url}/.well-known/jwks.json"
        self.access_token_url = f"{url}/api/v1/login/access-token"

    def get_public_key(
        self,
        *,
        key: t.Optional[str] = None,
        kid: str = "taktile-service",
    ) -> t.Any:
        jwk = (
            json.loads(key)
            if key
            else requests.get(self.public_key_url).json()
        )
        for k in jwk["keys"]:
            if k["kid"] == kid:
                return RSAAlgorithm.from_jwk(k)  # type: ignore
        raise InvalidAuthException("invalid-public-key")

    def decode_id_token(
        self,
        *,
        token: t.Optional[str] = None,
        key: t.Optional[str] = None,
        kid: str = "taktile-service",
    ) -> TaktileIdToken:
        if not token:
            raise InvalidAuthException("no-auth-provided")
        try:
            public_key = self.get_public_key(key=key, kid=kid)
            payload = jwt.decode(
                token,
                public_key,
                algorithms=[ALGORITHM],
                audience=settings.ENV,
            )
            token_data = TaktileIdToken(**payload)
        except jwt.ExpiredSignatureError:
            raise InvalidAuthException("signature-expired")
        except (jwt.PyJWTError, ValidationError):
            raise InvalidAuthException("could-not-validate")

        return token_data

    def get_access(
        self,
        *,
        session_state: SessionState,
        key: t.Optional[str] = None,
        kid: str = "taktile-service",
    ) -> t.Tuple[TaktileIdToken, SessionState]:
        if not session_state.api_key and not session_state.jwt:
            raise InvalidAuthException("no-auth-proved")
        if not session_state.jwt:
            res = requests.post(
                self.access_token_url,
                headers=session_state.to_auth_headers(),
            )
            res.raise_for_status()
            session_state.jwt = res.json()["id_token"]
        return (
            self.decode_id_token(token=session_state.jwt, key=key, kid=kid),
            session_state,
        )
